@@ -11,7 +11,7 @@
 /***********************************************************************/
 #include <my_global.h>
 #include <m_string.h>
-#if defined(_WIN32)
+#if defined(__WIN__)
 //nclude <io.h>
 //nclude <fcntl.h>
 #include <direct.h>                      // for getcwd
@@ -45,13 +45,13 @@
 #include "osutil.h"
 
 
-#if defined(_WIN32)
+#if defined(__WIN__)
 /***********************************************************************/
 /*  For dynamic load of ODBC32.DLL                                     */
 /***********************************************************************/
 #pragma comment(lib, "odbc32.lib")
 extern "C" HINSTANCE s_hModule;           // Saved module handle
-#endif // _WIN32
+#endif // __WIN__
 
 TYPCONV GetTypeConv();
 int GetConvSize();
@@ -125,17 +125,20 @@ int TranslateSQLType(int stp, int prec, int& len, char& v, bool& w)
   switch (stp) {
     case SQL_WVARCHAR:                      //  (-9)
       w = true;
+      /* fall through */
     case SQL_VARCHAR:                       //   12
       v = 'V';
       type = TYPE_STRING;
       break;
     case SQL_WCHAR:                         //  (-8)
       w = true;
+      /* fall through */
     case SQL_CHAR:                          //    1
       type = TYPE_STRING;
       break;
     case SQL_WLONGVARCHAR:                  // (-10)
       w = true;
+      /* fall through */
     case SQL_LONGVARCHAR:                   //  (-1)
 			if (GetTypeConv() == TPC_YES || GetTypeConv() == TPC_FORCE) {
 				v = 'V';
@@ -1280,19 +1283,16 @@ bool ODBConn::DriverConnect(DWORD Options)
   SWORD   nResult;
   PUCHAR  ConnOut = (PUCHAR)PlugSubAlloc(m_G, NULL, MAX_CONNECT_LEN);
   UWORD   wConnectOption = SQL_DRIVER_COMPLETE;
-#if defined(_WIN32)
+#if defined(__WIN__)
   HWND    hWndTop = GetForegroundWindow();
   HWND    hWnd = GetParent(hWndTop);
 
   if (hWnd == NULL)
     hWnd = GetDesktopWindow();
-#else   // !_WIN32
+#else   // !__WIN__
   HWND    hWnd = (HWND)1;
-#endif  // !_WIN32
-  PGLOBAL& g = m_G;
-  PDBUSER dup = PlgGetUser(g);
+#endif  // !__WIN__
 
-//if (Options & noOdbcDialog || dup->Remote)
     wConnectOption = SQL_DRIVER_NOPROMPT;
 //else if (Options & forceOdbcDialog)
 //  wConnectOption = SQL_DRIVER_PROMPT;
@@ -1301,10 +1301,10 @@ bool ODBConn::DriverConnect(DWORD Options)
                         SQL_NTS, ConnOut, MAX_CONNECT_LEN,
                         &nResult, wConnectOption);
 
-#if defined(_WIN32)
+#if defined(__WIN__)
   if (hWndTop)
     EnableWindow(hWndTop, true);
-#endif   // _WIN32
+#endif   // __WIN__
 
   // If user hit 'Cancel'
   if (rc == SQL_NO_DATA_FOUND) {
@@ -1686,7 +1686,7 @@ int ODBConn::PrepareSQL(char *sql)
     b = false;
 
     if (m_hstmt) {
-      RETCODE rc = SQLFreeStmt(m_hstmt, SQL_CLOSE);
+      SQLFreeStmt(m_hstmt, SQL_CLOSE);
 
       hstmt = m_hstmt;
       m_hstmt = NULL;
@@ -1694,7 +1694,7 @@ int ODBConn::PrepareSQL(char *sql)
       if (m_Tdb->GetAmType() != TYPE_AM_XDBC)
         ThrowDBX(MSG(SEQUENCE_ERROR));
 
-      } // endif m_hstmt
+    } // endif m_hstmt
 
     rc = SQLAllocStmt(m_hdbc, &hstmt);
 
@@ -2347,6 +2347,7 @@ int ODBConn::GetCatInfo(CATPARM *cap)
         break;
       case CAT_SPC:
         ThrowDBX("SQLSpecialColumns not available yet");
+        break;
       default:
         ThrowDBX("Invalid SQL function id");
       } // endswitch infotype
