@@ -1,5 +1,5 @@
 /************* tabbson C++ Program Source Code File (.CPP) *************/
-/* PROGRAM NAME: tabbson     Version 1.2                               */
+/* PROGRAM NAME: tabbson     Version 1.1                               */
 /*  (C) Copyright to the author Olivier BERTRAND          2020 - 2021  */
 /*  This program are the BSON class DB execution routines.             */
 /***********************************************************************/
@@ -53,7 +53,6 @@ USETEMP UseTemp(void);
 bool    JsonAllPath(void);
 int     GetDefaultDepth(void);
 char   *GetJsonNull(void);
-bool    Stringified(PCSZ, char*);
 
 /***********************************************************************/
 /* BSONColumns: construct the result blocks containing the description */
@@ -174,7 +173,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
   lvl = GetIntegerTableOption(g, topt, "Depth", lvl);
   sep = GetStringTableOption(g, topt, "Separator", ".");
   sz = GetIntegerTableOption(g, topt, "Jsize", 1024);
-  limit = GetIntegerTableOption(g, topt, "Limit", 50);
+  limit = GetIntegerTableOption(g, topt, "Limit", 10);
   strfy = GetStringTableOption(g, topt, "Stringify", NULL);
 
   /*********************************************************************/
@@ -194,11 +193,7 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
   if (!(tdp->Database = SetPath(g, db)))
     return 0;
 
-  if ((tdp->Objname = GetStringTableOption(g, topt, "Object", NULL))) {
-    if (*tdp->Objname == '$') tdp->Objname++;
-    if (*tdp->Objname == '.') tdp->Objname++;
-  } // endif Objname
-
+  tdp->Objname = GetStringTableOption(g, topt, "Object", NULL);
   tdp->Base = GetIntegerTableOption(g, topt, "Base", 0) ? 1 : 0;
   tdp->Pretty = GetIntegerTableOption(g, topt, "Pretty", 2);
   tdp->Xcol = GetStringTableOption(g, topt, "Expand", NULL);
@@ -223,7 +218,8 @@ int BSONDISC::GetColumns(PGLOBAL g, PCSZ db, PCSZ dsn, PTOS topt)
 
   if (tdp->Uri) {
 #if defined(JAVA_SUPPORT) || defined(CMGO_SUPPORT)
-    tdp->Collname = GetStringTableOption(g, topt, "Tabname", NULL);
+    tdp->Collname = GetStringTableOption(g, topt, "Name", NULL);
+    tdp->Collname = GetStringTableOption(g, topt, "Tabname", tdp->Collname);
     tdp->Schema = GetStringTableOption(g, topt, "Dbname", "test");
     tdp->Options = (PSZ)GetStringTableOption(g, topt, "Colist", "all");
     tdp->Pipe = GetBooleanTableOption(g, topt, "Pipeline", false);
@@ -437,7 +433,7 @@ bool BSONDISC::Find(PGLOBAL g, PBVAL jvp, PCSZ key, int j)
     jcol.Type = TYPE_UNKNOWN;
     jcol.Len = jcol.Scale = 0;
     jcol.Cbn = true;
-  } else  if (j < lvl && !Stringified(strfy, colname)) {
+  } else  if (j < lvl && !(strfy && !stricmp(strfy, colname))) {
     if (!fmt[bf])
       strcat(fmt, colname);
 
@@ -508,7 +504,7 @@ bool BSONDISC::Find(PGLOBAL g, PBVAL jvp, PCSZ key, int j)
     } // endswitch Type
 
   } else if (lvl >= 0) {
-    if (Stringified(strfy, colname)) {
+    if (strfy && !stricmp(strfy, colname)) {
       if (!fmt[bf])
         strcat(fmt, colname);
 
@@ -608,51 +604,33 @@ void BSONDISC::AddColumn(PGLOBAL g)
 /***********************************************************************/
 PBVAL BTUTIL::FindRow(PGLOBAL g)
 {
-  char *p, *objpath = PlugDup(g, Tp->Objname);
-  char *sep = (char*)(Tp->Sep == ':' ? ":[" : ".[");
-  bool  bp = false, b = false;
+  char *p, *objpath;
   PBVAL jsp = Tp->Row;
   PBVAL val = NULL;
 
-  for (; jsp && objpath; objpath = p, bp = b) {
-    if ((p = strpbrk(objpath + 1, sep))) {
-      b = (*p == '[');
+  for (objpath = PlugDup(g, Tp->Objname); jsp && objpath; objpath = p) {
+    if ((p = strchr(objpath, Tp->Sep)))
       *p++ = 0;
-    } // endif p
 
-    if (!bp && *objpath != '[' && !IsNum(objpath)) { // objpass is a key
+    if (*objpath != '[' && !IsNum(objpath)) { // objpass is a key
       val = (jsp->Type == TYPE_JOB) ?
         GetKeyValue(jsp, objpath) : NULL;
     } else {
-      if (bp || *objpath == '[') {                   // Old style
-        if (objpath[strlen(objpath) - 1] != ']') {
-          sprintf(g->Message, "Invalid Table path %s", Tp->Objname);
-          return NULL;
-        } else if (!bp)
+      if (*objpath == '[') {
+        if (objpath[strlen(objpath) - 1] == ']')
           objpath++;
-
-      } // endif bp
+        else
+          return NULL;
+      } // endif [
 
       val = (jsp->Type == TYPE_JAR) ?
-        GetArrayValue(jsp, atoi(objpath) - Tp->B) : NULL;
+        GetArrayValue(GetArray(jsp), atoi(objpath) - Tp->B) : NULL;
     } // endif objpath
 
       //  jsp = (val) ? val->GetJson() : NULL;
     jsp = val;
   } // endfor objpath
 
-  if (jsp && jsp->Type != TYPE_JOB) {
-    if (jsp->Type == TYPE_JAR) {
-      jsp = GetArrayValue(jsp, Tp->B);
-
-      if (jsp->Type != TYPE_JOB)
-        jsp = NULL;
-
-    } else
-      jsp = NULL;
-
-  } // endif Type
-
   return jsp;
 } // end of FindRow
 
@@ -676,22 +654,17 @@ PBVAL BTUTIL::MakeTopTree(PGLOBAL g, int type)
   if (Tp->Objname) {
     if (!Tp->Row) {
       // Parse and allocate Objpath item(s)
-      char *p, *objpath = PlugDup(g, Tp->Objname);
-      char *sep = (char*)(Tp->Sep == ':' ? ":[" : ".[");
+      char* p;
+      char *objpath = PlugDup(g, Tp->Objname);
       int   i;
-      bool  bp = false, b = false;
       PBVAL objp = NULL;
       PBVAL arp = NULL;
 
-      for (; objpath; objpath = p, bp = b) {
-        if ((p = strpbrk(objpath + 1, sep))) {
-          b = (*p == '[');
+      for (; objpath; objpath = p) {
+        if ((p = strchr(objpath, Tp->Sep)))
           *p++ = 0;
-        } // endif p
 
-
-        if (!bp && *objpath != '[' && !IsNum(objpath)) {
-          // objpass is a key
+        if (*objpath != '[' && !IsNum(objpath)) {
           objp = NewVal(TYPE_JOB);
 
           if (!top)
@@ -703,15 +676,15 @@ PBVAL BTUTIL::MakeTopTree(PGLOBAL g, int type)
           val = NewVal();
           SetKeyValue(objp, MOF(val), objpath);
         } else {
-          if (bp || *objpath == '[') {
+          if (*objpath == '[') {
             // Old style
             if (objpath[strlen(objpath) - 1] != ']') {
               sprintf(g->Message, "Invalid Table path %s", Tp->Objname);
               return NULL;
-            } else if (!bp)
+            } else
               objpath++;
 
-          } // endif bp
+          } // endif objpath
 
           if (!top)
             top = NewVal(TYPE_JAR);
@@ -783,16 +756,10 @@ void BCUTIL::SetJsonValue(PGLOBAL g, PVAL vp, PBVAL jvp)
           break;
         case TYPE_DATE:
           if (jvp->Type == TYPE_STRG) {
-            PSZ dat = GetString(jvp);
-
-            if (!IsNum(dat)) {
-              if (!((DTVAL*)vp)->IsFormatted())
-                ((DTVAL*)vp)->SetFormat(g, "YYYY-MM-DDThh:mm:ssZ", 20, 0);
-
-              vp->SetValue_psz(dat);
-            } else
-              vp->SetValue(atoi(dat));
+            if (!((DTVAL*)vp)->IsFormatted())
+              ((DTVAL*)vp)->SetFormat(g, "YYYY-MM-DDThh:mm:ssZ", 20, 0);
 
+            vp->SetValue_psz(GetString(jvp));
           } else
             vp->SetValue(GetInteger(jvp));
 
@@ -871,7 +838,7 @@ PBVAL BCUTIL::MakeBson(PGLOBAL g, PBVAL jsp, int n)
 /***********************************************************************/
 PBVAL BCUTIL::GetRowValue(PGLOBAL g, PBVAL row, int i)
 {
-  int    nod = Cp->Nod, n = nod - 1;
+  int    nod = Cp->Nod;
   JNODE *nodes = Cp->Nodes;
   PBVAL  arp;
   PBVAL  bvp = NULL;
@@ -1190,15 +1157,10 @@ bool BSONDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
   G = g;
   Schema = GetStringCatInfo(g, "DBname", Schema);
   Jmode = (JMODE)GetIntCatInfo("Jmode", MODE_OBJECT);
-
-  if ((Objname = GetStringCatInfo(g, "Object", NULL))) {
-    if (*Objname == '$') Objname++;
-    if (*Objname == '.') Objname++;
-  } // endif Objname
-
+  Objname = GetStringCatInfo(g, "Object", NULL);
   Xcol = GetStringCatInfo(g, "Expand", NULL);
   Pretty = GetIntCatInfo("Pretty", 2);
-  Limit = GetIntCatInfo("Limit", 50);
+  Limit = GetIntCatInfo("Limit", 10);
   Base = GetIntCatInfo("Base", 0) ? 1 : 0;
   Sep = *GetStringCatInfo(g, "Separator", ".");
   Accept = GetBoolCatInfo("Accept", false);
@@ -1209,7 +1171,7 @@ bool BSONDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
     Collname = GetStringCatInfo(g, "Name",
       (Catfunc & (FNC_TABLE | FNC_COL)) ? NULL : Name);
     Collname = GetStringCatInfo(g, "Tabname", Collname);
-    Options = GetStringCatInfo(g, "Colist", Xcol ? "all" : NULL);
+    Options = GetStringCatInfo(g, "Colist", NULL);
     Filter = GetStringCatInfo(g, "Filter", NULL);
     Pipe = GetBoolCatInfo("Pipeline", false);
     Driver = GetStringCatInfo(g, "Driver", NULL);
@@ -1253,7 +1215,7 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
 
     if (Lrecl) {
       // Allocate the parse work memory
-      G = PlugInit(NULL, (size_t)Lrecl * (Pretty < 0 ? 3 : 5));
+      G = PlugInit(NULL, (size_t)Lrecl * (Pretty < 0 ? 2 : 4));
     } else {
       strcpy(g->Message, "LRECL is not defined");
       return NULL;
@@ -1287,7 +1249,6 @@ PTDB BSONDEF::GetTable(PGLOBAL g, MODE m)
 #endif  // !MONGO_SUPPORT
       } // endif Driver
 
-      Pretty = 4;   // Not a file
     } else if (Zipped) {
 #if defined(ZIP_SUPPORT)
       if (m == MODE_READ || m == MODE_ANY || m == MODE_ALTER) {
@@ -1493,7 +1454,7 @@ int TDBBSN::EstimatedLength(void)
 } // end of Estimated Length
 
 /***********************************************************************/
-/*  OpenDB: Data Base open routine for BSN access method.              */
+/*  OpenDB: Data Base open routine for JSN access method.              */
 /***********************************************************************/
 bool TDBBSN::OpenDB(PGLOBAL g)
 {
@@ -1715,7 +1676,6 @@ BSONCOL::BSONCOL(PGLOBAL g, PCOLDEF cdp, PTDB tdbp, PCOL cprec, int i)
   Xpd = false;
   Parsed = false;
   Warned = false;
-  Sgfy = false;
 } // end of BSONCOL constructor
 
 /***********************************************************************/
@@ -1735,7 +1695,6 @@ BSONCOL::BSONCOL(BSONCOL* col1, PTDB tdbp) : DOSCOL(col1, tdbp)
   Xpd = col1->Xpd;
   Parsed = col1->Parsed;
   Warned = col1->Warned;
-  Sgfy = col1->Sgfy;
 } // end of BSONCOL copy constructor
 
 /***********************************************************************/
@@ -1974,10 +1933,6 @@ bool BSONCOL::ParseJpath(PGLOBAL g)
       // Analyse intermediate array processing
       if (SetArrayOptions(g, p, i, Nodes[i - 1].Key))
         return true;
-      else if (Xpd && Tbp->Mode == MODE_DELETE) {
-        strcpy(g->Message, "Cannot delete expanded columns");
-        return true;
-      } // endif Xpd
 
     } else if (*p == '*') {
       // Return JSON
@@ -2011,12 +1966,11 @@ PSZ BSONCOL::GetJpath(PGLOBAL g, bool proj)
       if (*p1 == '$') p1++;
       if (*p1 == '.') p1++;
       mgopath = PlugDup(g, p1);
-    } else {
-      Sgfy = true;
+    } else
       return NULL;
-    } // endif
 
     for (p1 = p2 = mgopath; *p1; p1++)
+    {
       if (i) {                 // Inside []
         if (isdigit(*p1)) {
           if (!proj)
@@ -2052,20 +2006,16 @@ PSZ BSONCOL::GetJpath(PGLOBAL g, bool proj)
       case '*':
         if (*(p2 - 1) == '.' && !*(p1 + 1)) {
           p2--;              // Suppress last :*
-          Sgfy = true;
           break;
         } // endif p2
-
+        /* fall through */
       default:
         *p2++ = *p1;
         break;
       } // endswitch p1;
-
-      if (*(p2 - 1) == '.')
-        p2--;
-
-      *p2 = 0;
-      return mgopath;
+    }
+    *p2 = 0;
+    return mgopath;
   } else
     return NULL;
 
@@ -2280,6 +2230,8 @@ int TDBBSON::MakeDocument(PGLOBAL g)
     return RC_FX;
 
   if ((objpath = PlugDup(g, Objname))) {
+    if (*objpath == '$') objpath++;
+    if (*objpath == '.') objpath++;
     p1 = (*objpath == '[') ? objpath++ : NULL;
 
     /*********************************************************************/
